package Atlas::Project::Contig;

$VERSION=1.0;

use strict;
use Atlas::Utility::ObjectAttribute;
use Atlas::Project::Trace;

use vars qw($VERBOSE);
local $VERBOSE=0;

my @_contig_attribute=(
		       'name',
		       'length',
		       'start',
		       'end',
		       'project'
		       );
&Atlas::Utility::ObjectAttribute::declare_rw(@_contig_attribute);


# A new contig is born.
sub create {
    my($class)=shift;
    
    my($self)=+{
	'name'=>'-none-',
	'length'=>0,
	'traces'=>[],
	'samples'=>{},
	'start'=>0,
	'end'=>0,
    };
    bless($self,$class);
}



# Add a reading to the Contig
# 
# name   contig_start contig_end  strand(+/-)
sub add_trace {
    my($self)=shift;
    my($trace)=shift;

    push(@{$self->{'traces'}},$trace);
    $trace->contig($self);

    # Add the trace to its sample group for the contig
    push(@{$self->{'samples'}{$trace->mapgap_group()}},$trace);

}

sub list_traces {
    my($self)=shift;

    @{$self->{'traces'}};
}

sub sort_traces_by_contig_start {
    my($self)=shift;

    @{$self->{'traces'}}=sort({$a->contig_start()<=>$b->contig_start()} 
			      @{$self->{'traces'}});
}

# Return a list of traces in a sample group.
sub list_traces_for_sample {
    my($self)=shift;
    my($sample)=shift;

    @{$self->{'samples'}{$sample}};
}

sub list_samples {
    my($self)=shift;

    sort(keys(%{$self->{'samples'}}));
}

sub list_pair_samples {
    my($self)=shift;

    if (exists $self->{'pair_samples'}) {
	return @{$self->{'pair_samples'}};
    } else {
	return;
    }
}


##############################################
#PURPOSE: Find out all pair end sequences in a contig and save them into
#    an array under the contig.
#Performance: essentially look at all traces in the contig.  The performance can
# be improved if combine with other action when you see one pair end, however, it also
# limits the reusability.
#
#
sub find_pair_samples {

    my($contig)=shift;
    my($samples, $current_template, $current_traces);

    $samples=$contig->{'samples'}; #get the ref of the sample hash.
        #
        # traverse the hash and save templates that are pair-sequenced.
        #

    foreach $current_template (keys %{$samples}) {

	$current_traces=$samples->{$current_template}; #get the ref of trace array.

             # traverse the trace array for each template
	if($#$current_traces < 1) { next;}
             
	my $i=0;
	if( $current_traces->[$i]->{'contig_strand'} eq '+' ) {
	    for(; $i <= $#$current_traces; $i++) {
		if($current_traces->[$i]->{'contig_strand'} eq '-') {
		    push(@{$contig->{'pair_samples'}}, $current_traces);
		    last;
		}
	    }
	}
        else {
	    for(; $i <= $#$current_traces; $i++) {
		if($current_traces->[$i]->{'contig_strand'} eq '+') {
		    push(@{$contig->{'pair_samples'}}, $current_traces);
		    last;
		} 
	    }
	}
    }
}    

=pod
=head1 ContainCloneEnd

    check if an assembled contig contains the cloning end of a give project, if yes, give the startpoint, side and orientation of it

        left <----  ---->right
    -----------------------------
   left          start          right

    $contig->ContainCloneEnd(\%endread, $vectorlen, $endlen, $baclib);

   the %endread contains all types of end read (vector internal read, junction read and bac end read), detail see clonend package.  $vectorlen is the length of the cloning vector, $endlen is the cloning vector end fragment length. 
   baclib is the regular expression for bac library

   sp6 ------------------------------------ T7 cloing vector orientation

  return: the hash contains all the traces that link the contig to the end, actually, the estimated position of
          cloning ends for the contig

=cut
sub ContainCloneEnd {

    my $self=shift;
    my ($endreadref, $vectorlen, $endlen, $baclib)=@_; 

    my %end_hash;  #a hash to rem all traces the link the contig to the end

    #get all the traces in the contig and scan all of them for end read
    my @contig_samples=$self->list_samples();
    my $contigname=$self->name();

  SKIP:
    foreach my $sample (@contig_samples) {
	my $vector_sample; #the sample name in the vector
	if(exists $endreadref->{$sample}) {
	    $vector_sample=$sample;
	    print "\n==================$sample\n\n" if($VERBOSE);
	} elsif($sample=~/_(\w+?)(\d+)$/ && exists $endreadref->{($1 . '_' . $2)}) {
	    $vector_sample=$1 . '_' . $2;
	    print "\n==================$sample\n\n" if($VERBOSE);
	} else {
	    next;
	}
	print $self->length() . "\n" if($VERBOSE);
	
	#get the right mapped vector read
	my $side; my $distance=$vectorlen;
	my $side_read; my $strand='+';
	foreach my $vector_read (@{$endreadref->{$vector_sample}}) { #one place to change back
	    if(! defined $vector_read->contig() ) { #an internal read
		if($vector_read->contig_strand eq '+') {
		    if( ! defined $side) {
			$side = 'T7';
		    } else {
			next SKIP unless ($side eq 'T7'); #there is conflict within the same sample
		    }
		    #only record the most close ones
		    if( ($vectorlen-$vector_read->contig_start()+1)<$distance ) {
			$distance=$vectorlen-$vector_read->contig_start()+1;
			$side_read=$vector_read;
		    }
		} else {
		    if( ! defined $side) {
			$side = 'SP6';
		    } else {
			next SKIP unless ($side eq 'SP6'); #there is conflict within the same sample
		    }
		    if( $vector_read->contig_end()<$distance ) {
			$distance=$vector_read->contig_end();
			$side_read=$vector_read;
		    }
		}
	    } else { #an end read
		if (! defined $side) {
		    $side = $vector_read->contig();	   
		} else {
		    next SKIP unless ($side eq $vector_read->contig() ); #there is conflict within the same sample
		}
		if($vector_read->contig_strand eq '+') {
		    $strand='+';
		    $distance=$endlen-$vector_read->contig_start()+1; 
                                                          #for bac end read, $vector_read->contig_start=$endlen
		} else {
		    $strand='-';
		    $distance=-1; #the clone is right on the junction and the other end is inside of the vector
		}
		$side_read=$vector_read;
	    }
	} #end of look though all vector read

	#look through all traces from the current sample
	my @traces=$self->list_traces_for_sample($sample);

	my ($start_point, $orientation); #start_point, the joint between vector and contig
                                         #orientation, +/- is NoFlip/Flip the contig regarding vector

	my $lib_name;
	$lib_name=$side_read->coslib() if(defined $side_read);
	my $insertsize=2000; #it is baylor reads, 2000 in average

	#print "The side_read is " . $side_read->original_name() ." " . $side_read->chem() ."\n";
	foreach my $trace (@traces) {
	    ###############################
	    if ($trace->original_name() eq $side_read->original_name()) { 
		next unless ( (defined $side_read->contig()) ); #it is an end read
		if( $strand eq '+') { #forward end read
		    if($trace->contig_strand eq '+') {
			$start_point= $trace->contig_start();
			$orientation='+'; #the contig is to the right of the clone end
		    } else {
			$start_point = $trace->contig_end();
			$orientation='-';
		    }
		} else { #the reverse end read
		    if($trace->contig_strand eq '+') {
			$orientation = '-';
			$start_point = $trace->contig_end();
		    } else {
			$orientation = '+';
			$start_point = $trace->contig_start();
		    }		    
		}
	    } else { #internal read
		#skip bac end read mate
		next if($trace->coslib()=~/$baclib/);
		if($trace->contig_strand eq '+') {
		    $orientation = '-';
		    $start_point = $trace->contig_start() + $insertsize - $distance;
		} else {
		    $orientation = '+';
		    $start_point= $trace->contig_end() - $insertsize + $distance;
		}
	    }
	    #print "orientation is $orientation\n";
	    push @{$end_hash{$side}->{$orientation}}, $start_point;  #save it to the contig hash
	} #end for each traces for the current sample
    } #end of all samples for the contig

    return \%end_hash if (%end_hash); #return it if not empty
}

#-------------------------------------------------------
#HELPER FUNCTIONS
#-------------------------------------------------------



1;

__END__
